/*
 * Soundgraph IMON IR/VFD USB device driver for FreeBSD 6.2 and higher
 *
 *   Copyright (C) 2007-2008 Alexander Popov
 *
 * Based on the Linux imon driver
 *  
 *   Copyright (C) 2004 Venky Raju(dev@venky.ws)
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */


/* TODO:
 * - Soundgraph Vendor id should be added to usbdevs.h
 * - (minor) use result == (0 or USB_NORMAL_COMPLETION) ?
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/ioccom.h>
#include <sys/fcntl.h>
#include <sys/filio.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/tty.h>
#include <sys/file.h>
#if __FreeBSD_version >= 500014
#include <sys/selinfo.h>
#else
#include <sys/select.h>
#endif
#include <sys/poll.h>
#include <sys/sysctl.h>
#include <sys/uio.h>

/* conditional variables */
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/condvar.h>

#include <dev/usb/usb.h>
#include <dev/usb/usb_port.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>

/* This is where USB vendor Ids are declared 
#include <dev/usb/usbdevs.h> */

#include "lirc.h"

/*#define IMON_DEBUG*/

#ifdef IMON_DEBUG
#define DPRINTF(args...) printf(args)
#else
#define DPRINTF(args...)
#endif

/* structure describes a particular USB endpoint */
struct imon_endpoint
{
   usbd_interface_handle      iface;
   usb_endpoint_descriptor_t* epdesc;
   usbd_pipe_handle           pipeh;
   u_char*                    ibuf;    /* start of buffer */
   struct clist               iqueue;  /* input queue */
   /* IR protocol decode data */
   int                   dec_count;    /* length of 0 or 1 sequence */
   int                   dec_prev_bit; /* logic level of sequence */
   int              dec_initial_space; /* initial space flag */
   
};

#define IMON_MAX_ENDPOINTS    2        /* maximum nr of supported endpoints */
#define VFD_ENDPOINT          0        /* endpoint (array) index for VFD */
#define IR_ENDPOINT           1        /* endpoint (array) index for IR  */

#define LIRC_DEVICE           "lirc%d" /* dev-node for IR */
#define VFD_DEVICE            "lcd%d"  /* dev-node for LCD */

#define IMONMINOR(u, e)       (((u) << 4) | (e))
#define IMONENDPOINT(n)       (minor(n) & 0xf)
#define IMONUNIT(n)           ((minor(n) >> 4) & 0xf)

struct imon_softc
{
   USBBASEDEVICE sc_dev;
   struct cdev*  ir_dev;
   struct cdev*  vfd_dev;

   usbd_device_handle udev;
   struct imon_endpoint endpoints[IMON_MAX_ENDPOINTS];
    
   char vfd_proto_6p;         /* VFD requires 6th packet */
   char ir_onboard_decode;    /* IR signals decoded onboard */
   struct mtx imon_mtx;       /* mutex to syncronize access to softsc data */
   int state;
#define IMON_INIT    0x00     /* initial driver state */
#define IMON_ASLEEP  0x01     /* user process is waiting for data */
#define IMON_DYING   0x02     /* device is being detached */
   int imon_refcnt;           /* reference counting driver use */
   struct cv imon_cv;         /* cond variable - user processes are gone */
};

Static d_open_t  imonopen;
Static d_close_t imonclose;
Static d_write_t imonwrite;
Static d_read_t  imonread;
Static d_ioctl_t imonioctl;

Static struct cdevsw imon_cdevsw = {
   .d_version =  D_VERSION,
   .d_flags   =  0,
   .d_open    =  imonopen,
   .d_close   =  imonclose,
   .d_write   =  imonwrite,
   .d_read    =  imonread,
   .d_ioctl   =  imonioctl,
   .d_name    =  "imon"
};

Static void imonintr(usbd_xfer_handle xfer, usbd_private_handle addr, 
                     usbd_status status);

/* Describes supported devices */
typedef struct imon_device_description
{
   u_int vendor_id;
   u_int product_id;
} imon_device_description_t;

static imon_device_description_t imon_devices[] = {
   { 0x0aa8, 0xffda },         /* IR & VFD    */
   { 0x0aa8, 0x8001 },         /* IR only     */
   { 0x15c2, 0xffda },         /* IR & VFD    */
   { 0x15c2, 0xffdc },         /* IR & VFD    */
   { 0x04e8, 0xff30 }          /* ext IR only */
};

/* Some IMON VFD models require 6th packet */
static u_int vfd_proto_6p_vendor_list[] = { 0x15c2 };

static unsigned char vfd_packet6 [] = {
		0x01, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF };

/* Newer IMON models decode the signal onboard */
static u_int ir_onboard_decode_product_list[] = { 0xffdc };

/* other definitions */
#define VFD_MAX_PAYLOAD_SIZE          32   /* maximum incoming data size */
#define VFD_TRANSFER_BUFFER_LENGTH    35   /* 5 packets, each 7 bytes */
#define VFD_TRANSFER_PACKET_SIZE      8    /* interrupt transfer takes 8 bytes, 
                                              7 carrying data and last one a 
                                              sequence number */
#define VFD_TRANSFER_DATA_CHUNK_SIZE  7    /* size of actual data transferred in
                                              one interrupt */
#define VFD_TRANSFER_SEQ_INCREMENT    2
#define VFD_FILLER_CHAR               ' '

#define IR_TRANSFER_PACKET_SIZE       8    /* expect to receive 8 bytes packets
                                              from remote control */
#define IR_BUFFER_SIZE                1020 /* buffer size */
#define IR_READ_BUFFER_SIZE           1024
#define BIT_DURATION                  250  /* each bit received is 250us */

/*------------------------------------------------------------------------------
   Forward declarations 
  ----------------------------------------------------------------------------*/

static void
init_softc(struct imon_softc* sc);
static void
filter_store_endpoint(struct imon_softc* sc, usbd_interface_handle iface,
                      int iendptno, usb_endpoint_descriptor_t *edesc);
static int
vfd_transfer(struct imon_softc* sc, usbd_xfer_handle xfer, 
             usbd_pipe_handle pipeh, unsigned char* buffer, int count);
static void
cleanup_softc(struct imon_softc* sc);
static void 
submit_data(struct imon_softc* sc);


USB_DECLARE_DRIVER(imon);

USB_MATCH(imon)
{
   DPRINTF("imon: USB_MATCH() - enter\n");

   USB_MATCH_START(imon, uaa);
   usb_device_descriptor_t* dev_descr = NULL;
   int result = UMATCH_NONE;
    
   if (!uaa->iface)
   {
       result = UMATCH_NONE;
   }
   else
   {
      dev_descr = usbd_get_device_descriptor(uaa->device);
      if (dev_descr != NULL)
      {
         DPRINTF("imon: checking idVendor = 0x%x, idProduct = 0x%x\n", 
                 UGETW(dev_descr->idVendor), UGETW(dev_descr->idProduct));

         int count = sizeof(imon_devices) / sizeof(imon_device_description_t);
         int found = 0;
         int i     = 0;
         for (i = 0; (i < count) && !found; i++)
         {
            if ((imon_devices[i].vendor_id == UGETW(dev_descr->idVendor)) && 
                (imon_devices[i].product_id == UGETW(dev_descr->idProduct)))
            {
               DPRINTF("imon: found a match, idVendor = 0x%x, idProduct = 0x%x\n",
                       UGETW(dev_descr->idVendor), UGETW(dev_descr->idProduct));
               found = 1;
            }
         }

         if (found)
         {
            result = UMATCH_VENDOR_PRODUCT;
         }
      }
   }

   DPRINTF("imon: USB_MATCH() - exit: %x\n", result);
   return (result);
}

USB_ATTACH(imon)
{
   USB_ATTACH_START(imon, sc, uaa);
   usbd_status result = USBD_NORMAL_COMPLETION;
   char devinfo[1024]; 
   usbd_devinfo(uaa->device, 0, devinfo);
   USB_ATTACH_SETUP;
 
   DPRINTF("imon: USB_ATTACH(device: %s) - enter\n", USBDEVNAME(sc->sc_dev));
    
   /* initialize elements of the sc structure */
   init_softc(sc);

   sc->udev = uaa->device;

   /* find out how many interfaces this device supports */
   u_int8_t intf_count = 0;

   if (result == USBD_NORMAL_COMPLETION)
   {
      result = usbd_interface_count(sc->udev, &intf_count);
      if (result == USBD_NORMAL_COMPLETION)
      {
         DPRINTF("imon: device supports %d interfaces\n", intf_count);
      }
      else
      {
         DPRINTF("imon: failed to retrieve interface count\n");
      }
   }
    
   /* this driver supports only one interface */
   int ifaceno = 0;

   /* retrieve interface */
   usbd_interface_handle iface = NULL;

   if (result == USBD_NORMAL_COMPLETION)
   {
      result = usbd_device2interface_handle(sc->udev, ifaceno, &iface);
      if (result != USBD_NORMAL_COMPLETION)
      {
         DPRINTF("imon: failed to retrieve interface handle\n");
      }
   }

   /* find out number of endpoints */
   u_int8_t endpoint_count = 0;

   if (result == USBD_NORMAL_COMPLETION)
   {
      result = usbd_endpoint_count(iface, &endpoint_count);
      if (result == USBD_NORMAL_COMPLETION)
      {
         DPRINTF("imon: device supports %d endpoints\n", endpoint_count);
      }
      else
      {
         DPRINTF("imon: failed to retrieve endpoint count\n");
      }
   }

   /* Filter out and store required endpoints */
   if (result == USBD_NORMAL_COMPLETION)
   {
      int iendptno = 0;
      usb_endpoint_descriptor_t *edesc = NULL;
      for (iendptno = 0; iendptno < endpoint_count; iendptno++)
      {
         edesc = usbd_interface2endpoint_descriptor(iface, iendptno);
         filter_store_endpoint(sc, iface, iendptno, edesc);
      }
   }
    
   /* Determine if VFD requires 6th packet */

   usb_device_descriptor_t* dev_descr = NULL;
   if (result == USBD_NORMAL_COMPLETION)
   {
      dev_descr = usbd_get_device_descriptor(sc->udev);
      if (dev_descr != NULL)
      {
         int count = sizeof(vfd_proto_6p_vendor_list) / 
                     sizeof(vfd_proto_6p_vendor_list[0]);
         int found = 0;
         int i     = 0;
         for (i = 0; (i < count) && !found; i++)
         {
            if (vfd_proto_6p_vendor_list[i] == UGETW(dev_descr->idVendor))
            {
               /* Requires 6th packet */
               sc->vfd_proto_6p = 1;
               found = 1;
            }
         }
      }
   }

   DPRINTF("imon: VFD requires 6th packet: %d\n", sc->vfd_proto_6p);
    
   /* Determine if the IR signals are decoded onboard */

   if (result == USBD_NORMAL_COMPLETION)
   {
      dev_descr = usbd_get_device_descriptor(sc->udev);
      if (dev_descr != NULL)
      {
         int count = sizeof(ir_onboard_decode_product_list) / 
                     sizeof(ir_onboard_decode_product_list[0]);
         int found = 0;
         int i     = 0;
         for (i = 0; (i < count) && !found; i++)
         {
            if (ir_onboard_decode_product_list[i] == UGETW(dev_descr->idProduct))
            {
               /* Decoded onboard */
               sc->ir_onboard_decode = 1;
               found = 1;
            }
         }
      }
   }
    
   DPRINTF("imon: IR signals decoded onboard: %d\n", sc->ir_onboard_decode);
            
   /* IR: make device node */
    
   if ((result == USBD_NORMAL_COMPLETION) &&
       (sc->endpoints[IR_ENDPOINT].epdesc != NULL))
   {
      sc->ir_dev = make_dev(&imon_cdevsw, 
                            IMONMINOR(USBDEVUNIT(sc->sc_dev), IR_ENDPOINT),
                            UID_ROOT, GID_OPERATOR, 0644,
                            LIRC_DEVICE, USBDEVUNIT(sc->sc_dev));
                             
      if (sc->ir_dev != NULL)
      {
         DPRINTF("imon: successfully created device node for IR\n");
      }
      else
      {
         DPRINTF("imon: failed to create device node for IR\n");
         result = ENXIO;
      }
   }
    
   /* VFD: make device node */
   
   if ((result == USBD_NORMAL_COMPLETION) &&
       (sc->endpoints[VFD_ENDPOINT].epdesc != NULL))
   {
      sc->vfd_dev = make_dev(&imon_cdevsw, 
                             IMONMINOR(USBDEVUNIT(sc->sc_dev), VFD_ENDPOINT),
                             UID_ROOT, GID_OPERATOR, 0644,
                             VFD_DEVICE, USBDEVUNIT(sc->sc_dev));
                             
      if (sc->vfd_dev != NULL)
      {
         DPRINTF("imon: successfully created device node for VFD\n");
      }
      else
      {
         DPRINTF("imon: failed to create device node for VFD\n");
         result = ENXIO;
      }
   }
   
   DPRINTF("imon: USB_ATTACH() - exit: %x\n", result);
   return (result);
}


USB_DETACH(imon)
{
   USB_DETACH_START(imon, sc);
   usbd_status result = USBD_NORMAL_COMPLETION;
   int i = 0;
    
   DPRINTF("imon: USB_DETACH(device: %s) - enter\n", USBDEVNAME(sc->sc_dev));

   /* synchronize access to shared data */
   mtx_lock_spin(&sc->imon_mtx);
  
   /* close all open pipes */
   if (result == USBD_NORMAL_COMPLETION)
   {
      for (i = 0; i < IMON_MAX_ENDPOINTS; i++)
      {
         if (sc->endpoints[i].pipeh != NULL)
         {
            usbd_abort_pipe(sc->endpoints[i].pipeh);
            usbd_close_pipe(sc->endpoints[i].pipeh);
            sc->endpoints[i].pipeh = NULL;
         }
      }
   }

   /* notify user processes that are blocked on IO - we're going down */
   sc->state &= IMON_DYING;
   struct imon_endpoint* imon_ep = &(sc->endpoints[IR_ENDPOINT]);

   /* unlock softc */
   mtx_unlock_spin(&sc->imon_mtx);

   wakeup(imon_ep);
   
   DPRINTF("imon: waiting for processes to go away\n");
   mtx_lock_spin(&sc->imon_mtx);
   if (sc->imon_refcnt <= 0)
   {
      DPRINTF("imon: there are no processes to wait for\n");
   }
   else
   {
      /*int wait_result = cv_timedwait_sig(&sc->imon_cv, 
                                         &sc->imon_mtx, hz * 10);
      DPRINTF("imon: woke, status: %x, refcnt: %d\n", 
              wait_result, sc->imon_refcnt);
      */
   }
   mtx_unlock_spin(&sc->imon_mtx);
   DPRINTF("imon: finished waiting for processes\n");
   
   /* destroy device nodes */
    
   if ((result == USBD_NORMAL_COMPLETION) && (sc->ir_dev != NULL))
   {
      DPRINTF("imon: destroying IR device node\n");
      destroy_dev(sc->ir_dev);
   }
   if ((result == USBD_NORMAL_COMPLETION) && (sc->vfd_dev != NULL))
   {
      DPRINTF("imon: destroying VFD device node\n");
      destroy_dev(sc->vfd_dev);
   }
    
   /* cleanup softc structure */
   cleanup_softc(sc);
    
   DPRINTF("imon: USB_DETACH() - exit: %x\n", result);
   return (result);
}

int imonopen(struct cdev* dev, int flag, int mode, usb_proc_ptr p)
{
   struct imon_softc* sc = NULL;
   int unit = IMONUNIT(dev);
	int iendptno = IMONENDPOINT(dev);
   USB_GET_SC_OPEN(imon, unit, sc);
   usbd_status status = USBD_NORMAL_COMPLETION;
   int result = 0;

   DPRINTF("imon: imonopen(u: %d, e: %d) - enter\n", unit, iendptno);

   /* synchronize access to shared data */
   mtx_lock_spin(&sc->imon_mtx);
   
   /* check driver state */
   if (result == 0)
   {
      if (sc->state & IMON_DYING)
      {
         DPRINTF("imon: driver state is dying\n");
         result = EIO;
      }
   }
   
   /* check if this endpoint has already been opened */
   if (result == 0)
   {
      if (sc->endpoints[iendptno].pipeh != NULL)
      {
         DPRINTF("imon: a pipe to endpoint %d is already open\n", iendptno);
         result = EBUSY;
      }
   }
   
   /* VFD part */
   if ((result == 0) && (iendptno == VFD_ENDPOINT) &&
       (sc->endpoints[VFD_ENDPOINT].epdesc != NULL))
   {
      /* open pipe */
      status = usbd_open_pipe(sc->endpoints[VFD_ENDPOINT].iface, 
                              sc->endpoints[VFD_ENDPOINT].epdesc->bEndpointAddress, 
                              0, 
                              &(sc->endpoints[VFD_ENDPOINT].pipeh));
                                    
      if (status == USBD_NORMAL_COMPLETION)
      {
         DPRINTF("imon: successfully opened a pipe to VFD\n");
      }
      else
      {
         DPRINTF("imon: failed to open a pipe to VFD\n");
         result = EIO;
      }
   }
      
   /* IR part */
   if ((result == 0) && (iendptno == IR_ENDPOINT) &&
       (sc->endpoints[IR_ENDPOINT].epdesc != NULL))
   {
      struct imon_endpoint* imon_ep = &(sc->endpoints[IR_ENDPOINT]);
      int isize = UGETW(imon_ep->epdesc->wMaxPacketSize);
      imon_ep->ibuf = malloc(isize, M_USBDEV, M_WAITOK);
      (void)clalloc(&(imon_ep->iqueue), IR_BUFFER_SIZE, 0);
      
      status = usbd_open_pipe_intr(imon_ep->iface,
                                   imon_ep->epdesc->bEndpointAddress,
                                   USBD_SHORT_XFER_OK,
                                   &(imon_ep->pipeh),
                                   sc, /* so that intr handler knows */
                                   imon_ep->ibuf,
                                   isize,
                                   imonintr,
                                   USBD_DEFAULT_INTERVAL);
      if (status == USBD_NORMAL_COMPLETION)
      {
         DPRINTF("imon: successfully opened a pipe to IR\n");
      }
      else
      {
         DPRINTF("imon: failed to open a pipe to IR\n");
         /* clean up */
         free(imon_ep->ibuf, M_USBDEV); 
         imon_ep->ibuf = NULL;
         clfree(&(imon_ep->iqueue));
         result = EIO;
      }
   }
   
   /* increase reference count */
   if (result == 0)
   {
      sc->imon_refcnt++;
   }

   /* unlock softc */
   mtx_unlock_spin(&sc->imon_mtx);

   DPRINTF("imon: imonopen(u: %d, e: %d) - exit: %x (%d)\n", 
           unit, iendptno, result, sc->imon_refcnt);
   return (result);
}

int imonclose(struct cdev* dev, int flag, int mode, usb_proc_ptr p)
{
   struct imon_softc* sc = NULL;
   int unit = IMONUNIT(dev);
	int iendptno = IMONENDPOINT(dev);
   USB_GET_SC(imon, unit, sc);
   int result = 0;

   DPRINTF("imon: imonclose(u: %d, e: %d) - enter\n", unit, iendptno);

   /* synchronize access to shared data */
   mtx_lock_spin(&sc->imon_mtx);
   
   if ((result == 0) && (sc->endpoints[iendptno].pipeh != NULL))
   {
      usbd_abort_pipe(sc->endpoints[iendptno].pipeh);
      usbd_close_pipe(sc->endpoints[iendptno].pipeh);
      sc->endpoints[iendptno].pipeh = NULL;
   }
   
   /* free buffers allocated for IR */
   if ((result == 0) && (iendptno == IR_ENDPOINT))
   {
      struct imon_endpoint* imon_ep = &(sc->endpoints[IR_ENDPOINT]);
      ndflush(&(imon_ep->iqueue), imon_ep->iqueue.c_cc);
      clfree(&(imon_ep->iqueue));
      if (imon_ep->ibuf != NULL)
      {
         free(imon_ep->ibuf, M_USBDEV);
         imon_ep->ibuf = NULL;
      }
   }

   /* decrease reference count */
   sc->imon_refcnt--;
   if (sc->imon_refcnt <= 0)
   {
      /* signal all user processes have gone */
      DPRINTF("imon: user processes have all gone\n");
      cv_signal(&sc->imon_cv);
      sc->imon_refcnt = 0;
   }
   
   /* unlock softc */
   mtx_unlock_spin(&sc->imon_mtx);

   DPRINTF("imon: imonclose(u: %d, e: %d) - exit: %x (%d)\n", 
           unit, iendptno, result, sc->imon_refcnt);
   return (result);
}


/* The IMON VFD is 2x16 characters and requires data in 5 consecutive USB 
   interrupt packets, each packet but the last carrying 7 bytes */
int imonwrite(struct cdev* dev, struct uio* uio, int flags)
{
   struct imon_softc* sc = NULL;
   int unit = IMONUNIT(dev);
	int iendptno = IMONENDPOINT(dev);
   USB_GET_SC(imon, unit, sc);
   int result = 0;
    
   DPRINTF("imon: imonwrite(u: %d, e: %d) - enter\n", unit, iendptno);
   
   /* synchronize access to shared data */
   mtx_lock_spin(&sc->imon_mtx);
   
   /* check driver state */
   if (result == 0)
   {
      if (sc->state & IMON_DYING)
      {
         DPRINTF("imon: driver state is dying\n");
         result = EIO;
      }
   }
   
   if (result == 0)
   {
      /* write is only allowed to VFD */
      if (iendptno != VFD_ENDPOINT)
      {
         DPRINTF("imon: write is only allowed to VFD\n");
         result = EIO;
      }
   }
   
   if (result == 0)
   {
      if (sc->endpoints[VFD_ENDPOINT].pipeh == NULL)
      {
         DPRINTF("imon: a pipe to VFD has not been opened\n");
         result = EIO;
      }
   }
       
   if (result == 0)
   {
      if ((uio->uio_iov->iov_len <= 0) || 
          (uio->uio_iov->iov_len > VFD_MAX_PAYLOAD_SIZE))
      {
         DPRINTF("imon: payload size exceeds maximum allowed\n");
         result = EIO;
      }
   }

   /* contains data to be transferred */
   unsigned char transferbuf[VFD_TRANSFER_BUFFER_LENGTH] = {0,};
        
   if (result == 0)
   {
      /* copy data from user land */
      result = copyin(uio->uio_iov->iov_base, transferbuf, 
                      uio->uio_iov->iov_len);
   }
    
   int i = 0;
   if (result == 0)
   {
      /* if less than VFD_MAX_PAYLOAD_SIZE bytes were sent, 
         fill in with spaces */
      for (i = uio->uio_iov->iov_len; i < VFD_MAX_PAYLOAD_SIZE; i++)
      {
         transferbuf[i] = VFD_FILLER_CHAR;
      }
        
      /* fill in remaining bytes with 0xFF */
      for (i = VFD_MAX_PAYLOAD_SIZE; i < VFD_TRANSFER_BUFFER_LENGTH; i++)
      {
         transferbuf[i] = 0xFF;
      }
   }
    
   /* transfer handle */
   usbd_xfer_handle xfer = NULL;
    
   if (result == 0)
   {
      /* setup transfer */
      xfer = usbd_alloc_xfer(sc->udev);
      if (xfer == NULL)
      {
         DPRINTF("imon: could not setup the transfer\n");
         result = EIO;
      }
   }
    
   /* data for each interrupt transfer */
   unsigned char txbuf[VFD_TRANSFER_PACKET_SIZE] = {0,};
   int txcount = VFD_TRANSFER_PACKET_SIZE;
   int offset = 0;
   int seqnr  = 0;
    
   if (result == 0)
   {
      do
      {
         /* fill in current transfer buffer with data chunk */
         memcpy(txbuf, &(transferbuf[offset]), VFD_TRANSFER_DATA_CHUNK_SIZE);
         txbuf[VFD_TRANSFER_DATA_CHUNK_SIZE] = (unsigned char)seqnr;

         /* transfer packet */
         result = vfd_transfer(sc, xfer, sc->endpoints[VFD_ENDPOINT].pipeh, 
                               txbuf, txcount);
         if (result == 0)
         {
            /* go for next data chunk */
            offset += VFD_TRANSFER_DATA_CHUNK_SIZE;
            seqnr  += VFD_TRANSFER_SEQ_INCREMENT;
         }
      } 
      while ((offset < VFD_TRANSFER_BUFFER_LENGTH) && 
             (result == 0));
   }
    
   if (result == 0)
   {
      /* send 6th packet if required */
      if (sc->vfd_proto_6p)
      {
         /* fill in current transfer buffer with 6th packet */
         memcpy(txbuf, vfd_packet6, VFD_TRANSFER_DATA_CHUNK_SIZE);
         txbuf[VFD_TRANSFER_DATA_CHUNK_SIZE] = (unsigned char)seqnr;

         /* transfer packet */
         result = vfd_transfer(sc, xfer, sc->endpoints[VFD_ENDPOINT].pipeh, 
                               txbuf, txcount);
      }
   }
    
   if (xfer != NULL)
   {
      /* free transfer handle */
      usbd_free_xfer(xfer);
   }
            
   /* unlock softc */
   mtx_unlock_spin(&sc->imon_mtx);
   
   DPRINTF("imon: imonwrite(u: %d, e: %d) - exit: %x\n", 
           unit, iendptno, result);
   return (result);
}

int imonread(struct cdev* dev, struct uio* uio, int flags)
{
   struct imon_softc* sc = NULL;
   int unit = IMONUNIT(dev);
	int iendptno = IMONENDPOINT(dev);
   USB_GET_SC(imon, unit, sc);
   int result = 0;
    
   DPRINTF("imon: imonread(u: %d, e: %d) - enter\n", unit, iendptno);
   
   /* synchronize access to shared data */
   mtx_lock_spin(&sc->imon_mtx);
   
   /* check driver state */
   if (result == 0)
   {
      if (sc->state & IMON_DYING)
      {
         DPRINTF("imon: driver state is dying\n");
         result = EIO;
      }
   }
   
   if (result == 0)
   {
      /* read is only allowed from IR */
      if (iendptno != IR_ENDPOINT)
      {
         DPRINTF("imon: reading allowed from IR only\n");
         result = EIO;
      }
   }
   
   if (result == 0)
   {
      if (sc->endpoints[IR_ENDPOINT].pipeh == NULL)
      {
         DPRINTF("imon: a pipe to IR has not been opened\n");
         result = EIO;
      }
   }
   
   struct imon_endpoint* imon_ep = &(sc->endpoints[IR_ENDPOINT]);
   if (result == 0)
   {
      if ((imon_ep->iqueue.c_cc == 0) && (flags & O_NONBLOCK))
      {
         DPRINTF("imon: no data and would block\n");
         result = EWOULDBLOCK;
      }
   }
   
   int wait_result = 0;
   if (result == 0)
   {
      while ((imon_ep->iqueue.c_cc == 0) && (result == 0))
      {
         DPRINTF("imon: imonread() - going to sleep\n");
         sc->state |= IMON_ASLEEP;
         wait_result = msleep_spin(imon_ep, &sc->imon_mtx, 
                                   "imonri",
                                   0 /* no timeout */);
         sc->state &= ~IMON_ASLEEP;
         DPRINTF("imon: imonread() - woke, status: %x\n", wait_result);
         result = (wait_result == 0) ? 0 : ENODEV;
         /* check driver state again */
         if (sc->state & IMON_DYING)
         {
            DPRINTF("imon: device state is dying\n");
            result = ENODEV;
         }
      }
   }
   
   /* transfer data */
   char buffer[IR_READ_BUFFER_SIZE]; /* do not do = {0,} !*/
   if (result == 0)
   {
      while ((imon_ep->iqueue.c_cc > 0) && (uio->uio_resid > 0) && 
             (result == 0))
      {
         u_int32_t n = min(imon_ep->iqueue.c_cc, uio->uio_resid);
         if (n > sizeof(buffer))
         {
            n = sizeof(buffer);
         }

         /* remove a small chunk from the input queue */
         q_to_b(&(imon_ep->iqueue), buffer, n);
         DPRINTF("imon: transferred %d bytes\n", n);

         /* copy the data to the user process */
         result = uiomove(buffer, n, uio);
      }
   }
   
   /* unlock softc */
   mtx_unlock_spin(&sc->imon_mtx);
   
   DPRINTF("imon: imonread(u: %d, e: %d) - exit: %x\n", 
           unit, iendptno, result);
   return (result);
}

int imonioctl(struct cdev* dev, u_long cmd, caddr_t data,
              int flag, usb_proc_ptr p)
{
   struct imon_softc* sc = NULL;
   int unit = IMONUNIT(dev);
	int iendptno = IMONENDPOINT(dev);
   USB_GET_SC(imon, unit, sc);
   int result = 0;
    
   DPRINTF("imon: imonioctl(u: %d, e: %d, cmd: %08lx) - enter\n", 
           unit, iendptno, cmd);
   
   /* ioctl is only implemented for IR */        
   if (result == 0)
   {
      if (iendptno != IR_ENDPOINT)
      {
         DPRINTF("imon: ioctl is only implemented for IR - exit\n");
         /* must return ok here, otherwise there will be problems */
         return (0);
      }
   }
   
   if (result == 0)
   {
      if (sc->endpoints[IR_ENDPOINT].pipeh == NULL)
      {
         DPRINTF("imon: a pipe to IR has not been opened\n");
         result = EIO;
      }
   }
   
   unsigned long features = 0;
   if (result == 0)
   {
      features = (sc->ir_onboard_decode) ? 
                                 LIRC_CAN_REC_LIRCCODE : LIRC_CAN_REC_MODE2;
   }
   
   if (result == 0)
   {
      switch (cmd)
      {
         case LIRC_GET_FEATURES:
         {
            DPRINTF("imon: cmd = LIRC_GET_FEATURES\n");
            *(unsigned long*)data = features;
         }
         break;
         
         case LIRC_GET_REC_MODE:
         {
            DPRINTF("imon: cmd = LIRC_GET_REC_MODE\n");
            if (!(features & LIRC_CAN_REC_MASK))
            {
               result = ENODEV;
            }
            else
            {
               *(unsigned long*)data = 
                  LIRC_REC2MODE(features & LIRC_CAN_REC_MASK);
            }
         }
         break;
         
         case LIRC_SET_REC_MODE:
         {
            DPRINTF("imon: cmd = LIRC_SET_REC_MODE\n");
            if (!(features & LIRC_CAN_REC_MASK))
            {
               result = ENODEV;
            }
            else if (!(LIRC_MODE2REC((*(unsigned long*)data)) & features))
            {
               result = EINVAL;
            }
         }
         break;
         
         case LIRC_GET_LENGTH:
         {
            DPRINTF("imon: cmd = LIRC_GET_LENGTH\n");
            *(unsigned long*)data = 
               (sc->ir_onboard_decode) ? 32 : (sizeof(lirc_t) * 8);
         }
         break;
         
         default:
         {
            DPRINTF("imon: cmd = other\n");
            result = ENOTTY;
         }
      }
   }
   
   DPRINTF("imon: imonioctl(u: %d, e: %d, cmd: %08lx) - exit: %x\n", 
           unit, iendptno, cmd, result);
   return (result);
}


/* ------------------------------------------------------------------
   Helper functions 
   ------------------------------------------------------------------
*/


static void
init_softc(struct imon_softc* sc)
{
   /* don't do this:
   sc->sc_dev            = NULL;
   */
   sc->ir_dev            = NULL;
   sc->vfd_dev           = NULL;
   sc->udev              = NULL;
   sc->vfd_proto_6p      = 0;
   sc->ir_onboard_decode = 0;
   sc->state             = IMON_INIT;
   sc->imon_refcnt       = 0;

   /* initialize endpoints */
   int i = 0;
   for (i = 0; i < IMON_MAX_ENDPOINTS; i++)
   {
      sc->endpoints[i].iface  = NULL;
      sc->endpoints[i].epdesc = NULL;
      sc->endpoints[i].pipeh  = NULL;
      sc->endpoints[i].ibuf   = NULL;
      /* initial IR protocol decode variables */
      sc->endpoints[i].dec_count         = 0;
      sc->endpoints[i].dec_initial_space = 1;
      sc->endpoints[i].dec_prev_bit      = 0;
   }

   /* init mutex */
   mtx_init(&sc->imon_mtx, "imon-mtx", NULL, MTX_SPIN);
   /* init conditional variables */
   cv_init(&sc->imon_cv, "imon-cv");
}

static void
cleanup_softc(struct imon_softc* sc)
{
   /* destroy synchronization primitives */
   mtx_destroy(&sc->imon_mtx);
   cv_destroy(&sc->imon_cv);
   /* back to initial state */
   sc->state = IMON_INIT;
}

/* Filter given endpoint and store it in softc if it is IR and/or 
   VFD endpoint */
static void
filter_store_endpoint(struct imon_softc* sc, usbd_interface_handle iface,
                      int iendptno, usb_endpoint_descriptor_t *edesc)
{
   /* trace endpoint parameters */

   DPRINTF("imon: filter(): found endpoint nr %d\n"
           "Address:\t0x%x\n"
           "Direction:\t%s\n"
           "Transfer:\t0x%x\n"
           "         \t(0x00 - ctrl, 0x01 - iso, 0x02 - bulk, 0x03 - intr)\n"
           "PacketSize:\t%d\n"
           "Interval:\t%d\n",
           iendptno,
           edesc->bEndpointAddress,
           (UE_GET_DIR(edesc->bEndpointAddress) == UE_DIR_IN ? "IN" : "OUT"),
           (edesc->bmAttributes & UE_XFERTYPE),
           UGETW(edesc->wMaxPacketSize),
           edesc->bInterval);
    
   /* Filter & store */

   if ((UE_GET_DIR(edesc->bEndpointAddress) == UE_DIR_IN) &&
       ((edesc->bmAttributes & UE_XFERTYPE) == UE_INTERRUPT))
   {
      DPRINTF("imon: filter(): using endpoint %d (addr: 0x%x) for IR output\n",
              iendptno, edesc->bEndpointAddress);

      /* store */
      sc->endpoints[IR_ENDPOINT].iface  = iface;
      sc->endpoints[IR_ENDPOINT].epdesc = edesc;
   }
   else if ((UE_GET_DIR(edesc->bEndpointAddress) == UE_DIR_OUT) &&
            ((edesc->bmAttributes & UE_XFERTYPE) == UE_INTERRUPT))
   {
      DPRINTF("imon: filter(): using endpoint %d (addr: 0x%x) for VFD input\n",
              iendptno, edesc->bEndpointAddress);

      /* store */
      sc->endpoints[VFD_ENDPOINT].iface  = iface;
      sc->endpoints[VFD_ENDPOINT].epdesc = edesc;
   }
   else
   {
      DPRINTF("imon: filter(): ignoring endpoint %d (addr: 0x%x)\n",
              iendptno, edesc->bEndpointAddress);
   }
}


/* Helper function to perform an interrupt transfer. Also makes conversion of
   error types and reporting.
*/
static int
vfd_transfer(struct imon_softc* sc, usbd_xfer_handle xfer, 
             usbd_pipe_handle pipeh, unsigned char* buffer, int count)
{
   usbd_status status = USBD_NORMAL_COMPLETION;
   int result  = 0;
   int txcount = count;

   DPRINTF("imon: vfd_transfer(device: %s) - enter\n", USBDEVNAME(sc->sc_dev));

   /* transfer packet */
   status = usbd_intr_transfer(xfer, pipeh, 0,
                               USBD_DEFAULT_TIMEOUT,
                               buffer, &txcount, "imonwi");

   DPRINTF("imon: transferred %d bytes\n", txcount);

   if (status != USBD_NORMAL_COMPLETION)
   {
      if (status == USBD_INTERRUPTED)
      {
         DPRINTF("imon: transfer has been interrupted\n");
         result = EINTR;
      }
      else if (status == USBD_TIMEOUT)
      {
         DPRINTF("imon: transfer has timed out\n");
         result = ETIMEDOUT;
      }
      else
      {
         DPRINTF("imon: transfer has failed\n");
         result = EIO;
      }
   }

   DPRINTF("imon: vfd_transfer() - exit\n");
   return (result);
}

Static void 
imonintr(usbd_xfer_handle xfer, usbd_private_handle addr, usbd_status status)
{
   struct imon_softc* sc = addr;
   struct imon_endpoint* imon_ep = &(sc->endpoints[IR_ENDPOINT]);
   u_char* buf = imon_ep->ibuf;
   
   /* synchronize access to shared data */
   /* mtx_lock_spin(&sc->imon_mtx); */
   
   /* check driver state */
   if (status == USBD_NORMAL_COMPLETION)
   {
      if (sc->state & IMON_DYING)
      {
         DPRINTF("imon: imonintr() - driver state is dying\n");
         status = ENODEV;
      }
   }
   
   if (status != USBD_NORMAL_COMPLETION)
   {
      if (status == USBD_CANCELLED)
      {
         DPRINTF("imon: imonintr() - status = cancelled\n");
      }
      else
      {
         DPRINTF("imon: imonintr() - status = %d\n", status);
         usbd_clear_endpoint_stall_async(imon_ep->pipeh);
      }
   }

   /* check how many bytes have actually been transfered */   
   u_int32_t count = 0;
   if (status == USBD_NORMAL_COMPLETION)
   {
      usbd_get_xfer_status(xfer, NULL, NULL, &count, NULL);
      if (count != IR_TRANSFER_PACKET_SIZE)
      {
         DPRINTF("imon: imonintr() - invalid incoming packet size: %d\n", 
                 count);
         status = EIO;
      }
   }
   
   /* check if its not a filler frame */
   if (status == USBD_NORMAL_COMPLETION)
   {
      if (buf[7] == 0xFF)
      {
         /* filler frame, no data here */
         status = EIO;
      }
      else if ((buf[0] == 0xFF) &&
               (buf[1] == 0xFF) &&
               (buf[2] == 0xFF) &&
               (buf[3] == 0xFF) &&
               (buf[4] == 0xFF) &&
               (buf[5] == 0xFF) &&                    /* iMON 2.4G */
               ((buf[6] == 0x4E && buf[7] == 0xAF) || /* LT */
                (buf[6] == 0x5E && buf[7] == 0xAF)))  /* DT */
      {
         /* also a filler frame, no data here */
         status = EIO;
      }
   }

   // a bit of debug information
#ifdef IMON_DEBUG
   if (status == USBD_NORMAL_COMPLETION)
   {
      DPRINTF("imon: imonintr() - data =");
      for (int i = 0; i < IR_TRANSFER_PACKET_SIZE; i++)
      {
         DPRINTF(" %02x", buf[i]);
      }
      DPRINTF("\n");
   }
#endif

   if (status == USBD_NORMAL_COMPLETION)
   {
      if (sc->ir_onboard_decode)
      {
         /* The signals have been decoded onboard the iMON controller */
         (void)b_to_q(buf, count, &(imon_ep->iqueue));
         if (sc->state & IMON_ASLEEP)
         {
            DPRINTF("imon: imonintr() - waking processes waiting for data\n");
            wakeup(imon_ep);
         }
      }
      else
      {
         /* Translate received data to pulse and space lengths.
          * Received data is active low, i.e. pulses are 0 and
          * spaces are 1.
          *
          * My original algorithm was essentially similar to
          * Changwoo Ryu's with the exception that he switched
          * the incoming bits to active high and also fed an
          * initial space to LIRC at the start of a new sequence
          * if the previous bit was a pulse.
          *
          * I've decided to adopt his algorithm. */

         int chunk_num = buf[7];
         if ((chunk_num == 1) && (imon_ep->dec_initial_space))
         {
            /* LIRC requires a leading space */
            imon_ep->dec_prev_bit = 0;
            imon_ep->dec_count = 4;
            submit_data(sc);
            imon_ep->dec_count = 0;
         }
         
         int octet = 0;
         int bit = 0;
         
         for (octet = 0; octet < 5; ++octet)
         {
            u_char mask = 0x80;
            for (bit = 0; bit < 8; ++bit)
            {
               int curr_bit = !(buf[octet] & mask);
               if (curr_bit != imon_ep->dec_prev_bit)
               {
                  if (imon_ep->dec_count)
                  {
                     submit_data(sc);
                     imon_ep->dec_count = 0;
                  }
                  imon_ep->dec_prev_bit = curr_bit;
               }
               ++imon_ep->dec_count;
               mask >>= 1;
            }
         }
         
         if (chunk_num == 10)
         {
            if (imon_ep->dec_count)
            {
               submit_data(sc);
               imon_ep->dec_count = 0;
            }
            imon_ep->dec_initial_space = imon_ep->dec_prev_bit;
         }
      }
   }
   
   /* unlock softc */
   /* mtx_unlock_spin(&sc->imon_mtx); */
}   

/**
 * Convert bit count to time duration(in us) and submit the value to lirc
 */
static void submit_data(struct imon_softc* sc)
{
   DPRINTF("imon: submit_data() - enter\n");
   
   struct imon_endpoint* imon_ep = &(sc->endpoints[IR_ENDPOINT]);
   u_char buf[4];
   int value = imon_ep->dec_count;
   int i = 0;

   value *= BIT_DURATION;
   value &= PULSE_MASK;
   if (imon_ep->dec_prev_bit)
   {
      value |= PULSE_BIT;
   }
   
   for (i = 0; i < 4; ++i)
   {
      buf[i] = value >> (i * 8);
   }

   (void)b_to_q(buf, sizeof(buf), &(imon_ep->iqueue));
   if (sc->state & IMON_ASLEEP)
   {
      DPRINTF("imon: submit_data() - waking processes waiting for data\n");
      wakeup(imon_ep);
   }
   
   DPRINTF("imon: submit_data() - exit\n");
}

DRIVER_MODULE(imon, uhub, imon_driver, imon_devclass, usbd_driver_load, 0);
